from vision import *
import comtypes
from comtypes import COMError
from comtypes.client import GetBestInterface, GetActiveObject
import winUser
from logHandler import log
try:
	import _winreg as winreg
except:
	import winreg
import winUser
import Queue
import queueHandler
import api
import wx
import addonHandler
from locationHelper import RectLTRB
from operator import attrgetter
from config import conf
from extensionPoints import Action
addonHandler.initTranslation()

COM_CLASS='ZoomText.Application'
WINDOW_CLASSES=[
	u'ZT12MainWindow',
	u'ZT11MainWindow',
]

class ZoomText(Magnifier):
	name = "zoomtext"
	conflictingRoles = frozenset([ROLE_COLORENHANCER, ROLE_HIGHLIGHTER])
	description = "ZoomText"
	supportedTrackingContexts = (
		CONTEXT_CARET,
		CONTEXT_REVIEW,
		CONTEXT_NAVIGATOR,
	)

	@classmethod
	def _get_isInstalled(cls):
		try:
			with winreg.OpenKey(winreg.HKEY_CLASSES_ROOT,r"TypeLib\{9FBBC98E-AE78-4A1C-B635-1ECA2FBFFB40}\1.2\0\win32", 0, winreg.KEY_READ | winreg.KEY_WOW64_64KEY) as k:
				path=winreg.QueryValueEx(k,"")[0][:-3]
			comtypes.GUID.from_progid(COM_CLASS)
			return True
		except WindowsError:
			return False

	def _get_running(self):
		for cls in WINDOW_CLASSES:
			try:
				hwnd=winUser.FindWindow(cls,'')
				return True
			except WindowsError:
				pass
		return False

	def _get_connected(self):
		return bool(self.com)

	def _get_isMagnifying(self):
		return self.connected and self.magnificationLevel > 1.0

	def __init__(self, *roles):
		if not self.activeRoles:
			self.internalRect = None
			self.xFreeze = False
			self._com=None
			self.rect=None
			self.initialCapsState = conf['keyboard']['useCapsLockAsNVDAModifierKey']
		super(ZoomText, self).__init__(*roles)

	def initializeMagnifier(self):
		super(ZoomText, self).initializeMagnifier()

	def terminateMagnifier(self):
		super(ZoomText, self).terminateMagnifier()

	def clean(self):
		if self._com:
			self._com=None
		if self.rect:
			self.rect=None
		if self.initialCapsState:
			conf['keyboard']['useCapsLockAsNVDAModifierKey']=True

	def terminate(self, *roles):
		super(ZoomText, self).terminate(*roles)
		if self.activeRoles:
			return
		self.clean()

	def _get_com(self):
		if self._com:
			if self.running:
				return self._com
			else:
				self.clean()
				return None
		elif self.running:
			try:
				self._com=GetActiveObject(COM_CLASS, dynamic=True)
			except WindowsError:
				return None
			# An IRectangle object we can update continously to avoid garbage creation
			if not self.rect:
				self.rect=GetBestInterface(self._threadCom.CreateObject("Rectangle"))
			if self.initialCapsState:
				conf['keyboard']['useCapsLockAsNVDAModifierKey']=False
			return self._com

	def executeCommand(self,cmd):
		if self.connected:
			try:
				self.com.Hotkeys.Execute(cmd)
				if cmd.endswith("Magnification"):
					return self.magnificationLevel
				return True
			except COMError:
				log.debugWarning("Could not execute {cmd}".format(cmd=cmd), exc_info=True)
		return False

	def _set_magnificationLevel(self,level):
		if level<1.0:
			raise ValueError("Unable to set ZoomText magnifier level, value too low")
		elif level>60.0:
			raise ValueError("Unable to set ZoomText magnifier level, value too high")
		else:
			self.setVar('Magnification.PrimaryWindow.Power.level',level)

	def _get_magnificationLevel(self):
		return self.getVar('Magnification.PrimaryWindow.Power.level')

	def setVar(self,param,val):
		if not self.connected:
			raise RuntimeError("Unable to set %s"%param)
		split=param.rsplit('.',1)
		if len(split)>1:
			obj,attr=split
			attrgetter(obj+"__setattr__")(self.com)(attr,val)
		else:
			self.com.__setattr__(param,val)

	def getVar(self,param):
		if not self.connected:
			raise RuntimeError("Unable to get %s"%param)
		return attrgetter(param)(self.com)

	def toggleVar(self,toggle, vals=None):
		if not self.connected:
			raise RuntimeError("Unable to toggle %s"%toggle)
		state=self.getVar(toggle)
		if isinstance(state,bool):
			newState=not state
		elif isinstance(vals,list):
			try:
				index=vals.index(state)+1
			except ValueError:
				index=0
			newState=vals[index] if index<len(vals) else vals[0]
		else:
			raise NotImplementedError(state)
		self.setVar(toggle,newState)
		return newState

	def trackToRectangle(self, rect, context=CONTEXT_UNDETERMINED, area=None):
		if not self.connected:
			return
		if self.xFreeze:
			rect = RectLTRB(self.internalRect.left, rect.top, self.internalRect.right, rect.bottom)
		ZTRect=self.NVDARectToZTRect(rect,self.rect)
		self.com.Magnification.Tracking.TrackTo(ZTRect)
		self.internalRect = rect

	def NVDARectToZTRect(self,NVDARect,ZTRect=None):
		"""
		Converts a NVDA rectangle to a ZoomText rectangle.
		@param NVDARect: NVDA rectangle.
		@type NVDARect: L{locationHelper.RectLTRB}
		@param ZTRect: a ZoomText rectangle com object. If provided, only changes the coordinates on the provided object.
		@type ZTRect: pointer to L{IRectangle}
		@returns: a ZoomText rectangle com object.
		@rtype: pointer to L{IRectangle}
		"""
		if self.connected:
			if not ZTRect:
				ZTRect=GetBestInterface(self.com.CreateObject("Rectangle"))
			ZTRect.Set(*NVDARect)
			return ZTRect

	def ZTRectToNVDARect(self,ZTRect):
		"""
		Converts a ZoomText rectangle to a NVDA rectangle.
		@param rect:  a ZoomText rectangle com object.
		@type rect: pointer to L{IRectangle}
		@returns: NVDA rectangle.
		@rtype: L{locationHelper.RectLTRB}
		"""
		return RectLTRB(ZTRect.left,ZTRect.top,ZTRect.right,ZTRect.bottom)

	def _CommandScript(self,gesture, cmd, desc, doc, verbose=False):
		if self.connected:
			res = self.executeCommand(cmd)
			if not verbose:
				return
			if res:
				if cmd.endswith("Magnification"):
					msg=_("Magnification %.2f")%res
				else:
					msg=desc
			else:
				msg=_("{action} failed").format(action=desc)
		else:
			msg=_("{action} failed, magnifier not active").format(action=desc)
		ui.message(msg)
