from vision import *
import libUSA
from ctypes import *
from ctypes.wintypes import *
import wx
from scriptHandler import script
import ui
from logHandler import log
from locationHelper import RectLTWH, RectLTRB
import baseObject
from itertools import cycle
from operator import gt, lt
import addonHandler
from config import conf, isInstalledCopy,getUserDefaultConfigPath
import gui
from .configSpec import configSpec
from .dialogs import USASettingsPanel
import winUser

#We need to initialize translation and localization support:
addonHandler.initTranslation()

class USA(Magnifier, ColorEnhancer, baseObject.ScriptableObject):
	name = "usa"
	description = "Babbage Ultimate Screen Access"
	supportedTrackingContexts = (
		CONTEXT_FOCUS,
		CONTEXT_FOREGROUND,
		CONTEXT_CARET,
		CONTEXT_REVIEW,
		CONTEXT_NAVIGATOR,
		CONTEXT_MOUSE,
	)
	contextBoundaries = {
		CONTEXT_UNDETERMINED: RectLTRB(0,0,100,100),
		CONTEXT_FOCUS: RectLTRB(25,25,75,75),
		CONTEXT_FOREGROUND: RectLTRB(40,0,60,0),
		CONTEXT_CARET: RectLTRB(10,10,90,90),
		CONTEXT_REVIEW: RectLTRB(0,0,100,100),
		CONTEXT_NAVIGATOR: RectLTRB(0,0,100,100),
		CONTEXT_MOUSE: RectLTRB(-1,-1,-1,-1),
	}
	scriptCategory = _("Ultimate Screen Access")
	guiPanelCls = USASettingsPanel

	def __init__(self, *roles):
		if not self.activeRoles:
			self.initializeConfig()
			self._trialExpired = False
			libUSA.Initialize()
			self.createWindows()
		super(USA, self).__init__(*roles)

	def initializeConfig(self):
		if not "usa" in conf['vision']:
			conf['vision']['usa']={}
		conf['vision']['usa'].spec.update(configSpec)

	def initializeMagnifier(self):
		super(USA, self).initializeMagnifier()
		self._internalRect = api.getDesktopObject().location.toLTRB()
		self.xFreeze = False
		self.cachedMagLevel = None
		self.magnificationLevel = conf['vision']['usa']['magnificationLevel']

	def initializeColorEnhancer(self):
		super(USA, self).initializeColorEnhancer()
		self._transformation = conf['vision']['usa']['colorScheme']

	def terminate(self, *roles):
		super(USA, self).terminate(*roles)
		if self.activeRoles:
			return
		self.destroyWindows()
		libUSA.Terminate()

	def _get__hostWindowHandle(self):
		try:
			return winUser.FindWindow(u"A4U_USA", None)
		except WindowsError:
			return 0

	def _get_running(self):
		handle = self._hostWindowHandle
		if not handle:
			return False
		isVisible = winUser.isWindowVisible(handle)
		if not isVisible and not self._trialExpired:
			self._trialExpired = True
			wx.CallAfter(
				gui.messageBox,
				parent=gui.mainFrame,
				# Translators: Title of the trial dialog.
				caption=_("Trial Expired"),
				# Translators: Message shown when running in trial mode.
				message=_(
					"Your 10 minutes trial of Ultimate Screen Access has expired.\n"
					"If this message is unexpected, please check whether you have a valid license.\n"
				),
				style=wx.OK | wx.ICON_WARNING
			)
		return isVisible

	def createWindows(self):
		for rect in (wx.Display(i).GetGeometry() for i in xrange(wx.Display.GetCount())):
			rect = RectLTWH(*rect).toRECT()
			hwnd = libUSA.CreateMagnifier(rect)
		self.bitmapSmoothing = self.bitmapSmoothing

	predefinedMagnificationLevels = (
		1.0, 1.25, 1.5, 1.75,
		2.0, 2.5, 3.0,
		4.0, 5.0, 6.5, 8.0,
		10.0, 12.5, 15.0, 17.5,
		20.0, 25.0, 30.0, 37.5,
		45.0, 52.5, 60.0, 70.0, 80.0
	)
	def _set_magnificationLevel(self, level):
		for hwnd in libUSA.magWindows:
			libUSA.SetMagnificationLevel(hwnd, level)
		conf['vision']['usa']['magnificationLevel'] = level

	def _get_magnificationLevel(self):
		return libUSA.GetMagnificationLevel(libUSA.magWindows.keys()[0])

	def _set_bitmapSmoothing(self, state):
		for hwnd in libUSA.magWindows:
			try:
				libUSA.SetBitmapSmoothing(hwnd, state)
			except WindowsError:
				log.error("Bitmap smoothing not available for window %d" % hwnd)
				return
		conf['vision']['usa']['bitmapSmoothing'] = state

	def _get_bitmapSmoothing(self):
		return conf['vision']['usa']['bitmapSmoothing']

	def _getAvailableTransformations(self):
		return [
			ColorTransformationInfo("default", _("True colors"), libUSA.MAGCOLOREFFECT((
				(1.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, 1.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, 1.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(0.0, 0.0, 0.0, 0.0, 1.0)
			))),
			ColorTransformationInfo("inverted", _("Inverted colors"), libUSA.MAGCOLOREFFECT((
				(-1.0, 0.0, 0.0, 0.0, 0.0),
				(0.0, -1.0, 0.0, 0.0, 0.0),
				(0.0, 0.0, -1.0, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(1.0, 1.0, 1.0, 0.0, 1.0)
			))),
			ColorTransformationInfo("gray", _("Gray scale"), libUSA.MAGCOLOREFFECT((
				(0.3, 0.3, 0.3, 0.0, 0.0),
				(0.6, 0.6, 0.6, 0.0, 0.0),
				(0.1, 0.1, 0.1, 0.0, 0.0),
				(0.0, 0.0, 0.0, 1.0, 0.0),
				(0.0, 0.0, 0.0, 0.0, 1.0)
			))),
			ColorTransformationInfo("custom", _("Custom"), libUSA.MAGCOLOREFFECT((
				tuple(conf['vision']['usa']['customColorScheme']['r']),
				tuple(conf['vision']['usa']['customColorScheme']['g']),
				tuple(conf['vision']['usa']['customColorScheme']['b']),
				tuple(conf['vision']['usa']['customColorScheme']['a']),
				tuple(conf['vision']['usa']['customColorScheme']['t'])
			))),
		]

	def _get_transformation(self):
		return self._transformation

	def _set_transformation(self, transformation):
		info = self.availableTransformations[transformation]
		for hwnd in libUSA.magWindows:
			libUSA.SetColorScheme(hwnd, info.value)
		self._transformation = 		conf['vision']['usa']['colorScheme'] = transformation

	def destroyWindows(self):
		for hwnd in list(libUSA.magWindows):
			libUSA.DestroyMagnifier(hwnd)

	def terminateMagnifier(self):
		self._internalRect = api.getDesktopObject().location.toLTRB()
		self.xFreeze = False
		self.cachedMagLevel = None
		super(USA, self).terminateMagnifier()

	def terminateColorEnhancer(self):
		self._transformation = None
		super(USA, self).terminateColorEnhancer()

	def trackToRectangle(self, rect, context=CONTEXT_UNDETERMINED, area=None):
		if not rect:
			log.warning("No rectangle, unexpected behavior", stack_info=True)
		if self.xFreeze and self._internalRect:
			rect = RectLTRB(self._internalRect.left, rect.top, self._internalRect.right, rect.bottom)
		self._internalRect = rect
		r = rect.toRECT()
		boundary = self.contextBoundaries.get(context, self.contextBoundaries[CONTEXT_UNDETERMINED]).toRECT()
		for hwnd in libUSA.magWindows:
			libUSA.SetTrackingRectangle(hwnd, r, boundary)

	@script(gesture="kb:NVDA+i")
	def script_reportOffsets(self, gesture):
		MagGetWindowSourceFuncType = WINFUNCTYPE(BOOL, HWND, POINTER(RECT))
		MagGetWindowSourceArgTypes = ((1, "hwnd"), (2, "rect"))
		GetWindowSource = MagGetWindowSourceFuncType(("MagGetWindowSource", windll.magnification), MagGetWindowSourceArgTypes)
		rect = GetWindowSource(libUSA.magWindows.keys()[0])
		left = rect.left
		top = rect.top
		ui.message(str((left, top)))

	def _setMagnificationLevelHelper(self, increase=True):
		if not ROLE_MAGNIFIER in self.activeRoles or self.cachedMagLevel:
			ui.message(_("Magnifier functionality currently not enabled"))
			return
		curLevel = self.magnificationLevel
		levels = (iter if increase else reversed)(self.predefinedMagnificationLevels)
		for level in levels:
			if (gt if increase else lt)(level, curLevel):
				break
		else:
			level = curLevel
		self.magnificationLevel = level
		ui.message(_("Magnification level %s") % level)

	@script(gesture="kb:windows+NVDA+-", description=_("Decreases the magnification level"))
	def script_decreaseMagnification(self, gesture):
		self._setMagnificationLevelHelper(increase=False)

	@script(gesture="kb:windows+NVDA+=", description=_("Increases the magnification level"))
	def script_increaseMagnification(self, gesture):
		self._setMagnificationLevelHelper(increase=True)

	@script(gesture="kb:windows+NVDA+Enter", description=_("Toggles the magnifier"))
	def script_toggleMagnifier(self, gesture):
		if not ROLE_MAGNIFIER in self.activeRoles:
			ui.message(_("Magnifier functionality currently not enabled"))
			return
		if self.cachedMagLevel:
			level = self.magnificationLevel = self.cachedMagLevel
			self.cachedMagLevel = None
			ui.message(_("Magnifier on, level %s") % level)
		else:
			self.cachedMagLevel = self.magnificationLevel
			self.magnificationLevel = 1.0
			ui.message(_("Magnifier off"))

	@script(gesture="kb:windows+NVDA+x", description=_('Enables or disables freezing of the x axis when moving the magnified area'))
	def script_toggleXFreeze(self,gesture):
		if not ROLE_MAGNIFIER in self.activeRoles or self.cachedMagLevel:
			ui.message(_("Magnifier functionality currently not enabled"))
			return
		self.xFreeze = not self.xFreeze
		stateMsg = _("on") if self.xFreeze else _("off")
		ui.message(_("Freeze x axis %s")%stateMsg)

	@script(gesture="kb:windows+NVDA+c", description=_('Cycles between the available color schemes'))
	def script_toggleColorScheme(self,gesture):
		if not ROLE_COLORENHANCER in self.activeRoles:
			ui.message(_("Color enhancement functionality currently not enabled"))
			return
		IDs = list(self.availableTransformations.keys())
		transformations = list(self.availableTransformations.values())
		try:
			index = IDs.index(self.transformation)
		except:
			index=0
		newIndex = (index+1) % len(IDs)
		new = transformations[newIndex]
		self.transformation = IDs[newIndex]
		# Translators: Reports the newly activated color scheme.
		# %s will be replaced with the scheme's name.
		ui.message(_("Color scheme %s") % new.name)

	def _scrollScriptHelper(self, gesture, xDir=0, yDir=0):
		if not ROLE_MAGNIFIER in self.activeRoles or self.cachedMagLevel:
			ui.message(_("Magnifier functionality currently not enabled"))
			return
		currentRect = self._internalRect
		screen = api.getDesktopObject().location
		magLevel = self.magnificationLevel
		xSize = int(round(screen.right/magLevel))
		ySize = int(round(screen.bottom/magLevel))
		left, top, right, bottom = currentRect
		if xDir and self.xFreeze:
			self.script_toggleXFreeze(gesture)
		if xDir>0:
			right = min(screen.right, currentRect.right+int(round(xDir*xSize/2)))
			left = right - xSize
		elif xDir<0:
			left = max(screen.left, currentRect.left+int(round(xDir*xSize/2)))
			right = left + xSize
		if yDir>0:
			bottom = min(screen.bottom, currentRect.bottom+int(round(yDir*ySize/2)))
			top = bottom - ySize
		elif yDir<0:
			top = max(screen.top, currentRect.top+int(round(yDir*ySize/2)))
			bottom = top + ySize
		newRect = RectLTRB(left, top, right, bottom)
		if newRect == currentRect:
			ui.message(_("Edge of screen"))
			return
		self.trackToRectangle(newRect)

	@script(gesture="kb:windows+NVDA+upArrow", description=_("Scroll magnification area up"))
	def script_scrollUp(self, gesture):
		self._scrollScriptHelper(gesture, yDir=-1)

	@script(gesture="kb:windows+NVDA+downArrow", description=_("Scroll magnification area down"))
	def script_scrollDown(self, gesture):
		self._scrollScriptHelper(gesture, yDir=1)

	@script(gesture="kb:windows+NVDA+leftArrow", description=_("Scroll magnification area left"))
	def script_scrollLeft(self, gesture):
		self._scrollScriptHelper(gesture, xDir=-1)

	@script(gesture="kb:windows+NVDA+rightArrow", description=_("Scroll magnification area right"))
	def script_scrollRight(self, gesture):
		self._scrollScriptHelper(gesture, xDir=1)
